# standard libraries
import asyncio
import contextlib
import threading
import typing
import unittest

# third party libraries
# None

# local libraries
from nion.ui import UserInterface
from nion.utils import Model
from nion.utils import Binding


@contextlib.contextmanager
def event_loop_context() -> typing.Iterator[asyncio.AbstractEventLoop]:
    old_event_loop = asyncio.get_event_loop_policy().get_event_loop()
    event_loop = asyncio.new_event_loop()
    asyncio.get_event_loop_policy().set_event_loop(event_loop)
    yield event_loop
    event_loop.stop()
    event_loop.run_forever()
    event_loop.close()
    asyncio.get_event_loop_policy().set_event_loop(old_event_loop)


class TestCanvasItemClass(unittest.TestCase):

    def setUp(self) -> None:
        pass

    def tearDown(self) -> None:
        pass

    def test_update_binding_from_thread(self) -> None:
        # tests that setting the source model on a thread updates the ui model properly using bindable property.
        with event_loop_context() as event_loop:
            source_model = Model.PropertyModel(0)
            ui_model = Model.PropertyModel(1)
            binding = Binding.PropertyBinding(source_model, "value")

            def update_ui(value: int) -> None:
                ui_model.value = value

            with contextlib.closing(UserInterface.BindablePropertyHelper[int](None, update_ui)) as binding_helper:
                binding_helper.bind_value(binding)

                def set_value_on_thread2(value: int) -> None:
                    source_model.value = value

                t = threading.Thread(target=set_value_on_thread2, args=(7,))
                t.start()
                t.join()
            self.assertEqual(7, source_model.value)
            self.assertNotEqual(ui_model.value, source_model.value)
            # run the event loop twice. when the source model changes on a thread, it calls run_coroutine_threadsafe
            # which queues the call to make the call. the first run puts the correct call in the queue. the second
            # run executes the call.
            event_loop.stop()
            event_loop.run_forever()
            event_loop.stop()
            event_loop.run_forever()
            self.assertEqual(7, source_model.value)
            self.assertEqual(ui_model.value, source_model.value)



"""
A library of custom widgets.
"""

from __future__ import annotations

# standard libraries
import abc
import gettext
import functools
import typing

# local libraries
from nion.ui import Bitmap
from nion.ui import CanvasItem
from nion.ui import DrawingContext
from nion.ui import ListCanvasItem
from nion.ui import UserInterface
from nion.utils import Binding
from nion.utils import Color
from nion.utils import Event
from nion.utils import Geometry
from nion.utils import Model
from nion.utils import Selection

if typing.TYPE_CHECKING:
    from nion.ui import Window


_ = gettext.gettext


class WidgetSource:
    def __init__(self, ui: UserInterface.UserInterface, widget: typing.Optional[UserInterface.Widget], canvas_item: typing.Optional[CanvasItem.AbstractCanvasItem]) -> None:
        self.__ui = ui
        self.__widget = widget
        self.__canvas_item = canvas_item

    @property
    def canvas_item(self) -> CanvasItem.AbstractCanvasItem:
        assert self.__canvas_item
        return self.__canvas_item

    @property
    def widget(self) -> UserInterface.Widget:
        if self.__widget:
            return self.__widget
        assert self.__canvas_item
        canvas_widget = self.__ui.create_canvas_widget()
        canvas_widget.canvas_item.add_canvas_item(self.__canvas_item)
        return canvas_widget


class BaseWidgetCanvasItemController:
    def __init__(self, ui: UserInterface.UserInterface) -> None:
        super().__init__()
        self.__ui = ui

    @property
    def ui(self) -> UserInterface.UserInterface:
        return self.__ui

    @property
    @abc.abstractmethod
    def widget_source(self) -> WidgetSource: ...


class PushButtonWidgetCanvasItemController(BaseWidgetCanvasItemController):
    def __init__(self, ui: UserInterface.UserInterface) -> None:
        super().__init__(ui)
        self.on_clicked: typing.Optional[typing.Callable[[], None]] = None
        self.on_size_changed: typing.Optional[typing.Callable[[Geometry.IntSize], None]] = None

    @abc.abstractmethod
    def set_text(self, value: typing.Optional[str]) -> None: ...

    @abc.abstractmethod
    def set_icon(self, bitmap: typing.Optional[Bitmap.BitmapOrArray]) -> None: ...

    @abc.abstractmethod
    def set_enabled(self, enabled: bool) -> None: ...

    @abc.abstractmethod
    def set_tool_tip(self, tool_tip: typing.Optional[str]) -> None: ...

    @abc.abstractmethod
    def set_background_color(self, background_color: typing.Optional[typing.Union[str, DrawingContext.LinearGradient]]) -> None: ...

    @abc.abstractmethod
    def set_border_color(self, border_color: typing.Optional[str]) -> None: ...


class WidgetCanvasItemControllerFactory(typing.Protocol):

    def create_push_button_widget_canvas_item_controller(self) -> PushButtonWidgetCanvasItemController: ...

    def create_tab_widget_canvas_item_controller(self) -> TabWidgetCanvasItemController: ...


class CompositeWidgetBehavior(UserInterface.WidgetBehavior):

    def __init__(self, content_widget: UserInterface.Widget) -> None:
        self.content_widget = content_widget
        self.on_context_menu_event: typing.Optional[typing.Callable[[int, int, int, int], bool]] = None
        self.on_focus_changed: typing.Optional[typing.Callable[[bool], None]] = None
        self.on_ui_activity: typing.Optional[typing.Callable[[], None]] = None

    # subclasses should override to clear their variables.
    # subclasses should NOT call Qt code to delete anything here... that is done by the Qt code
    def close(self) -> None:
        # the content will be closed automatically since content widget gets returned as a contained item.
        # self.content_widget.close()
        self.content_widget = typing.cast(typing.Any, None)

    def periodic(self) -> None:
        self.content_widget.periodic()

    @property
    def widget(self) -> UserInterface.Widget:
        return self.content_widget

    def _set_root_container(self, root_container: typing.Optional[UserInterface.Window]) -> None:
        self.content_widget._set_root_container(root_container)

    def _get_content_widget(self) -> typing.Optional[UserInterface.Widget]:
        return self.content_widget

    def _set_focused(self, focused: bool) -> None:
        pass

    def _set_enabled(self, enabled: bool) -> None:
        pass

    def _set_tool_tip(self, tool_tip: typing.Optional[str]) -> None:
        pass

    def _set_background_color(self, background_color: typing.Optional[typing.Union[str, DrawingContext.LinearGradient]]) -> None:
        pass

    def _set_border_color(self, border_color: typing.Optional[str]) -> None:
        pass

    @property
    def focused(self) -> bool:
        return self.content_widget.focused

    @focused.setter
    def focused(self, focused: bool) -> None:
        self.content_widget.focused = focused
        self._set_focused(focused)

    @property
    def does_retain_focus(self) -> bool:
        return self.content_widget.does_retain_focus

    @does_retain_focus.setter
    def does_retain_focus(self, value: bool) -> None:
        self.content_widget.does_retain_focus = value

    @property
    def visible(self) -> bool:
        return self.content_widget.visible

    @visible.setter
    def visible(self, visible: bool) -> None:
        self.content_widget.visible = visible

    @property
    def enabled(self) -> bool:
        return self.content_widget.enabled

    @enabled.setter
    def enabled(self, enabled: bool) -> None:
        self.content_widget.enabled = enabled
        self._set_enabled(enabled)

    @property
    def size(self) -> Geometry.IntSize:
        return self.content_widget.size

    @size.setter
    def size(self, size: Geometry.IntSize) -> None:
        self.content_widget.size = size

    @property
    def tool_tip(self) -> typing.Optional[str]:
        return self.content_widget.tool_tip

    @tool_tip.setter
    def tool_tip(self, tool_tip: typing.Optional[str]) -> None:
        self.content_widget.tool_tip = tool_tip
        self._set_tool_tip(tool_tip)

    def set_background_color(self, value: typing.Optional[typing.Union[str, DrawingContext.LinearGradient]]) -> None:
        self.content_widget.background_color = value
        self._set_background_color(value)

    def set_border_color(self, value: typing.Optional[str]) -> None:
        self.content_widget.border_color = value
        self._set_border_color(value)

    def drag(self, mime_data: UserInterface.MimeData, thumbnail: typing.Optional[Bitmap.Bitmap] = None,
             hot_spot_x: typing.Optional[int] = None, hot_spot_y: typing.Optional[int] = None,
             drag_finished_fn: typing.Optional[typing.Callable[[str], None]] = None) -> None:
            self.content_widget.drag(mime_data, thumbnail, hot_spot_x, hot_spot_y, drag_finished_fn)

    def map_to_global(self, p: Geometry.IntPoint) -> Geometry.IntPoint:
        return self.content_widget.map_to_global(p)

    def set_property(self, key: str, value: typing.Any) -> None:
        pass


class BasicPushButtonWidgetCanvasItemController(PushButtonWidgetCanvasItemController):
    def __init__(self, ui: UserInterface.UserInterface) -> None:
        super().__init__(ui)

        self.__text_button_canvas_item = CanvasItem.TextButtonCanvasItem()
        self.__text_button_canvas_item.background_color = "#f0f0f0"
        self.__text_button_canvas_item.border_color = "gray"
        self.__icon_button_canvas_item = CanvasItem.BitmapButtonCanvasItem(padding=Geometry.IntSize(4, 4))
        self.__icon_button_canvas_item.background_color = "#f0f0f0"
        self.__icon_button_canvas_item.border_color = "gray"
        self.__stack = CanvasItem.CanvasItemComposition()
        self.__stack.layout = CanvasItem.CanvasItemLayout()
        self.__stack.add_canvas_item(self.__text_button_canvas_item)
        self.__stack.add_canvas_item(self.__icon_button_canvas_item)

        def handle_clicked() -> None:
            if callable(self.on_clicked):
                self.on_clicked()

        self.__text_button_canvas_item.on_button_clicked = handle_clicked
        self.__icon_button_canvas_item.on_button_clicked = handle_clicked

    @property
    def widget_source(self) -> WidgetSource:
        return WidgetSource(self.ui, None, self.__stack)

    def set_text(self, value: typing.Optional[str]) -> None:
        self.__text_button_canvas_item.visible = True
        self.__icon_button_canvas_item.visible = False
        self.__text_button_canvas_item.text = value or str()
        self.__text_button_canvas_item.size_to_content(self.ui.get_font_metrics)
        self.__icon_button_canvas_item.bitmap = None
        self.__icon_button_canvas_item.size_to_content(self.ui.get_font_metrics)

        if callable(self.on_size_changed):
            self.on_size_changed(Geometry.IntSize(width=int(self.__text_button_canvas_item.sizing.preferred_width or 0),
                                                  height=int(self.__text_button_canvas_item.sizing.preferred_height or 0)))

    def set_icon(self, bitmap: typing.Optional[Bitmap.BitmapOrArray]) -> None:
        self.__text_button_canvas_item.visible = False
        self.__icon_button_canvas_item.visible = True
        self.__icon_button_canvas_item.bitmap = Bitmap.promote_bitmap(bitmap)
        self.__icon_button_canvas_item.size_to_content(self.ui.get_font_metrics)
        self.__text_button_canvas_item.text = str()
        self.__text_button_canvas_item.size_to_content(self.ui.get_font_metrics)

        if callable(self.on_size_changed):
            self.on_size_changed(Geometry.IntSize(width=int(self.__icon_button_canvas_item.sizing.preferred_width or 0),
                                                  height=int(self.__icon_button_canvas_item.sizing.preferred_height or 0)))

    def set_enabled(self, enabled: bool) -> None:
        self.__text_button_canvas_item.enabled = enabled
        self.__icon_button_canvas_item.enabled = enabled

    def set_tool_tip(self, tool_tip: typing.Optional[str]) -> None:
        self.__text_button_canvas_item.tool_tip = tool_tip
        self.__icon_button_canvas_item.tool_tip = tool_tip

    def set_background_color(self, background_color: typing.Optional[typing.Union[str, DrawingContext.LinearGradient]]) -> None:
        self.__text_button_canvas_item.background_color = background_color
        self.__icon_button_canvas_item.background_color = background_color

    def set_border_color(self, border_color: typing.Optional[str]) -> None:
        self.__text_button_canvas_item.border_color = border_color
        self.__icon_button_canvas_item.border_color = border_color


class TabWidgetCanvasItemController(BaseWidgetCanvasItemController):
    def __init__(self, ui: UserInterface.UserInterface) -> None:
        super().__init__(ui)
        self.on_tab_clicked: typing.Optional[typing.Callable[[int], None]] = None

    @abc.abstractmethod
    def get_selected_index(self) -> int: ...

    @abc.abstractmethod
    def set_selected_index(self, selected_index: int) -> None: ...

    @abc.abstractmethod
    def add(self, child: UserInterface.Widget, label: str) -> None: ...


class BasicTabWidgetCanvasItemController(TabWidgetCanvasItemController):

    def __init__(self, ui: UserInterface.UserInterface) -> None:
        super().__init__(ui)
        self.__ui = ui
        self.__calculate_button_height()
        self.__header_bottom_border_properties = CanvasItem.CellBorderProperties(Color.Color("gray"), "solid", 1.0)
        header_bottom_border = CanvasItem.CellBorder()
        header_bottom_border.border_bottom = self.__header_bottom_border_properties
        self.__column_widget = ui.create_column_widget()
        self.__button_row = CanvasItem.CanvasItemComposition()
        self.__button_row.layout = CanvasItem.CanvasItemRowLayout()
        button_height = self.__button_height
        header_row = ui.create_row_widget()
        header_row_canvas_item = CanvasItem.CanvasItemComposition()
        header_row_canvas_item.layout = CanvasItem.CanvasItemRowLayout()
        header_left_spacing = CanvasItem.EmptyCanvasItem(border=header_bottom_border)
        header_left_spacing.update_sizing(header_left_spacing.sizing.with_fixed_width(8).with_fixed_height(button_height))
        header_right_stretch = CanvasItem.EmptyCanvasItem(border=header_bottom_border)
        header_right_stretch.update_sizing(header_right_stretch.sizing.with_fixed_height(button_height))
        header_row_canvas_item.add_canvas_item(header_left_spacing)
        header_row_canvas_item.add_canvas_item(self.__button_row)
        header_row_canvas_item.add_canvas_item(header_right_stretch)
        header_row_sizing_properties = {
            "height": button_height,
            "size-policy-horizontal": "expanding"
        }
        header_row_widget = ui.create_canvas_widget(header_row_sizing_properties)
        header_row_widget.canvas_item.add_canvas_item(header_row_canvas_item)
        header_row.add(header_row_widget)
        self.__stack = ui.create_stack_widget()
        self.__column_widget.add(header_row)
        self.__column_widget.add_spacing(4)
        self.__column_widget.add(self.__stack)
        self.__button_canvas_items: typing.List[CanvasItem.TextButtonCanvasItem] = list()
        self.__selected_index = 0

    def __calculate_button_height(self) -> None:
        button_canvas_item = CanvasItem.TextButtonCanvasItem("X")
        button_canvas_item.padding = Geometry.IntSize(width=6, height=4)
        button_canvas_item.size_to_content(self.ui.get_font_metrics)
        self.__button_height = int(button_canvas_item.sizing.preferred_height or 0.0)

    @property
    def widget_source(self) -> WidgetSource:
        return WidgetSource(self.ui, self.__column_widget, None)

    def get_selected_index(self) -> int:
        return self.__selected_index

    def set_selected_index(self, selected_index: int) -> None:
        selected_index = max(0, min(selected_index, len(self.__button_canvas_items)))
        self.__stack.current_index = selected_index
        self.__update_styles(selected_index)
        self.__selected_index = selected_index

    def add(self, child: UserInterface.Widget, label: str) -> None:
        button_canvas_item = CanvasItem.TextButtonCanvasItem(label)
        button_canvas_item.padding = Geometry.IntSize(width=6, height=4)
        button_canvas_item.size_to_content(self.ui.get_font_metrics)
        self.__button_row.add_canvas_item(button_canvas_item)
        self.__stack.add(child)
        self.__button_canvas_items.append(button_canvas_item)
        self.__update_styles(self.__selected_index)

        def button_clicked(index: int) -> None:
            if callable(self.on_tab_clicked):
                self.on_tab_clicked(index)

        button_canvas_item.on_button_clicked = functools.partial(button_clicked, len(self.__button_canvas_items) - 1)

    def __update_styles(self, selected_index: int) -> None:
        for index, button_canvas_item in enumerate(self.__button_canvas_items):
            border = CanvasItem.CellBorder()
            if selected_index == index:
                border.border_top = CanvasItem.CellBorderProperties(Color.Color("gray"))
                border.border_left = CanvasItem.CellBorderProperties(Color.Color("gray"))
                button_canvas_item.background_color = None
            else:
                border.border_top = CanvasItem.CellBorderProperties(Color.Color("gray"))
                border.border_left = CanvasItem.CellBorderProperties(Color.Color("gray"))
                border.border_bottom = self.__header_bottom_border_properties
                drawing_context = DrawingContext.DrawingContext()
                gradient = drawing_context.create_linear_gradient(0, self.__button_height, 0, 0, 0, self.__button_height)
                gradient.add_color_stop(0.0, Color.Color("gray").to_color_with_alpha(0.00).color_str)
                gradient.add_color_stop(0.5, Color.Color("gray").to_color_with_alpha(0.33).color_str)
                gradient.add_color_stop(1.0, Color.Color("gray").to_color_with_alpha(0.25).color_str)
                button_canvas_item.background_color = gradient
            if index == len(self.__button_canvas_items) - 1:
                border.border_right = CanvasItem.CellBorderProperties(Color.Color("gray"))
            button_canvas_item.border = border


class BasicWidgetCanvasItemControllerFactory(WidgetCanvasItemControllerFactory):

    def __init__(self, ui: UserInterface.UserInterface) -> None:
        self.__ui = ui

    def create_push_button_widget_canvas_item_controller(self) -> PushButtonWidgetCanvasItemController:
        return BasicPushButtonWidgetCanvasItemController(self.__ui)

    def create_tab_widget_canvas_item_controller(self) -> TabWidgetCanvasItemController:
        return BasicTabWidgetCanvasItemController(self.__ui)


class PushButtonWidgetBehavior(CompositeWidgetBehavior):
    def __init__(self, ui: UserInterface.UserInterface) -> None:
        self.__canvas_widget = ui.create_canvas_widget()
        super().__init__(self.__canvas_widget)

        widget_canvas_item_factory = BasicWidgetCanvasItemControllerFactory(ui)

        self.__push_button_widget_canvas_item_controller = widget_canvas_item_factory.create_push_button_widget_canvas_item_controller()

        canvas_item = self.__push_button_widget_canvas_item_controller.widget_source.canvas_item

        self.__canvas_widget.canvas_item.add_canvas_item(canvas_item)

        self.__text: typing.Optional[str] = None
        self.__icon: typing.Optional[Bitmap.Bitmap] = None
        self.on_clicked: typing.Optional[typing.Callable[[], None]] = None

        def handle_clicked() -> None:
            if callable(self.on_clicked):
                self.on_clicked()

        def handle_size_changed(size: Geometry.IntSize) -> None:
            self.__canvas_widget.size = size

        self.__push_button_widget_canvas_item_controller.on_clicked = handle_clicked
        self.__push_button_widget_canvas_item_controller.on_size_changed = handle_size_changed

    @property
    def text(self) -> typing.Optional[str]:
        return self.__text

    @text.setter
    def text(self, value: typing.Optional[str]) -> None:
        self.__text = value
        self.__push_button_widget_canvas_item_controller.set_text(value)

    @property
    def icon(self) -> typing.Optional[Bitmap.Bitmap]:
        return self.__icon

    @icon.setter
    def icon(self, bitmap: typing.Optional[Bitmap.BitmapOrArray]) -> None:
        self.__icon = Bitmap.promote_bitmap(bitmap)
        self.__push_button_widget_canvas_item_controller.set_icon(self.__icon)

    def _set_enabled(self, enabled: bool) -> None:
        self.__push_button_widget_canvas_item_controller.set_enabled(enabled)

    def _set_tool_tip(self, tool_tip: typing.Optional[str]) -> None:
        self.__push_button_widget_canvas_item_controller.set_tool_tip(tool_tip)

    def _set_background_color(self, background_color: typing.Optional[typing.Union[str, DrawingContext.LinearGradient]]) -> None:
        self.__push_button_widget_canvas_item_controller.set_background_color(background_color)

    def _set_border_color(self, border_color: typing.Optional[str]) -> None:
        self.__push_button_widget_canvas_item_controller.set_border_color(border_color)


class TabWidgetBehavior(CompositeWidgetBehavior):  # not subclass of UserInterface.TabWidgetBehavior until mypy #4125 is available
    def __init__(self, ui: UserInterface.UserInterface) -> None:
        column_widget = ui.create_column_widget()
        super().__init__(column_widget)
        self.__current_index_model = Model.PropertyModel[int](-1)

        widget_canvas_item_factory = BasicWidgetCanvasItemControllerFactory(ui)

        self.__tab_widget_canvas_item_controller = widget_canvas_item_factory.create_tab_widget_canvas_item_controller()

        column_widget.add(self.__tab_widget_canvas_item_controller.widget_source.widget)

        def value_changed(value: typing.Optional[int]) -> None:
            self.__tab_widget_canvas_item_controller.set_selected_index(value or 0)
            if callable(self.on_current_index_changed):
                self.on_current_index_changed(self.__tab_widget_canvas_item_controller.get_selected_index())

        def handle_tab_clicked(selected_index: int) -> None:
            self.__current_index_model.value = selected_index

        self.__tab_widget_canvas_item_controller.on_tab_clicked = handle_tab_clicked

        self.__current_index_model.on_value_changed = value_changed
        self.on_current_index_changed: typing.Optional[typing.Callable[[int], None]] = None

    def close(self) -> None:
        self.__tab_widget_canvas_item_controller.on_tab_clicked = None
        self.__current_index_model.on_value_changed = None
        self.on_current_index_changed = None
        super().close()

    def add(self, child: UserInterface.Widget, label: str) -> None:
        self.__tab_widget_canvas_item_controller.add(child, label)

    def restore_state(self, tag: str) -> None:
        pass

    def save_state(self, tag: str) -> None:
        pass

    @property
    def current_index(self) -> int:
        return self.__current_index_model.value or 0

    @current_index.setter
    def current_index(self, index: int) -> None:
        self.__current_index_model.value = index


class SectionWidget(UserInterface.Widget):
    """A widget representing a twist down section.

    The section is composed of a title in bold and then content.
    """

    def __init__(self, ui: UserInterface.UserInterface, section_title: str, section: UserInterface.Widget, section_id: typing.Optional[str] = None) -> None:
        section_widget = ui.create_column_widget()
        super().__init__(CompositeWidgetBehavior(section_widget))

        section_title_row = ui.create_row_widget()

        twist_down_canvas_item = CanvasItem.TwistDownCanvasItem()

        twist_down_canvas_widget = ui.create_canvas_widget(properties={"height": 20, "width": 20})
        twist_down_canvas_widget.canvas_item.add_canvas_item(twist_down_canvas_item)

        section_title_label = ui.create_label_widget(section_title)
        section_title_label.text_font = "bold"

        section_title_row.add(twist_down_canvas_widget)
        section_title_row.add(section_title_label)
        section_title_row.add_stretch()
        section_widget.add(section_title_row)
        section_content_row = ui.create_row_widget()
        section_content_column = ui.create_column_widget()
        section_content_column.add_spacing(4)
        section_content_column.add(section)
        section_content_row.add_spacing(20)
        section_content_row.add(section_content_column)
        section_widget.add(section_content_row)
        section_widget.add_spacing(4)

        initialized = False  # small optimization to avoid writing persistent string during initialization

        def set_expanded(expanded: typing.Optional[bool]) -> None:
            twist_down_canvas_item.checked = expanded or False
            section_content_column.visible = expanded or False
            if initialized and section_id:
                ui.set_persistent_string(section_id, "true" if twist_down_canvas_item.checked else "false")

        def toggle() -> None:
            self.expanded = not twist_down_canvas_item.checked

        section_open = ui.get_persistent_string(section_id, "true") == "true" if section_id else True
        twist_down_canvas_item.checked = section_open
        section_content_column.visible = section_open
        twist_down_canvas_item.on_button_clicked = toggle

        self.section_title_row = section_title_row
        self.__twist_down_canvas_item = twist_down_canvas_item

        def set_title(value: typing.Optional[str]) -> None:
            section_title_label.text = str(value) if value is not None else str()

        self.__title_binding_helper = UserInterface.BindablePropertyHelper[typing.Optional[str]](None, set_title)
        self.__expanded_binding_helper = UserInterface.BindablePropertyHelper[typing.Optional[bool]](None, set_expanded)

        self.title = section_title
        self.expanded = True

        initialized = True

    def close(self) -> None:
        self.__title_binding_helper.close()
        self.__title_binding_helper = typing.cast(typing.Any, None)
        self.__expanded_binding_helper.close()
        self.__expanded_binding_helper = typing.cast(typing.Any, None)
        super().close()

    @property
    def title(self) -> str:
        return self.__title_binding_helper.value or str()

    @title.setter
    def title(self, value: str) -> None:
        self.__title_binding_helper.value = value

    @property
    def expanded(self) -> bool:
        return self.__expanded_binding_helper.value or False

    @expanded.setter
    def expanded(self, value: bool) -> None:
        self.__expanded_binding_helper.value = value or False

    def bind_title(self, binding: Binding.Binding) -> None:
        self.__title_binding_helper.bind_value(binding)

    def unbind_title(self) -> None:
        self.__title_binding_helper.unbind_value()

    def bind_expanded(self, binding: Binding.Binding) -> None:
        self.__expanded_binding_helper.bind_value(binding)

    def unbind_expanded(self) -> None:
        self.__expanded_binding_helper.unbind_value()


class ListCanvasItemDelegate(ListCanvasItem.ListCanvasItemDelegate):

    def __init__(self) -> None:
        self.__items: typing.List[typing.Any] = list()
        self.on_item_selected: typing.Optional[typing.Callable[[int], None]] = None
        self.on_cancel: typing.Optional[typing.Callable[[], None]] = None

    @property
    def items(self) -> typing.Sequence[typing.Any]:
        return self.__items

    @items.setter
    def items(self, value: typing.Sequence[typing.Any]) -> None:
        self.__items = list(value)

    @property
    def item_count(self) -> int:
        return len(self.__items)

    def key_pressed(self, key: UserInterface.Key) -> bool:
        if key.is_escape:
            if callable(self.on_cancel):
                self.on_cancel()
                # returning False here will allow window to handle the escape key.
                return False
        return False

    def item_selected(self, index: int) -> bool:
        if callable(self.on_item_selected):
            self.on_item_selected(index)
            return True
        return False

    def paint_item(self, drawing_context: DrawingContext.DrawingContext, display_item: typing.Any, rect: Geometry.IntRect, is_selected: bool) -> None:
        raise NotImplementedError()


class ListWidget(UserInterface.Widget):
    """A widget with a list in a scroll bar."""

    def __init__(self, ui: UserInterface.UserInterface, list_item_delegate: ListCanvasItem.ListCanvasItemDelegate, *,
                 items: typing.Optional[typing.Sequence[typing.Any]] = None,
                 selection_style: typing.Optional[Selection.Style] = None,
                 properties: typing.Optional[typing.Mapping[str, typing.Any]] = None,
                 selection: typing.Optional[Selection.IndexedSelection] = None,
                 border_color: typing.Optional[str] = None, v_scroll_enabled: bool = True,
                 v_auto_resize: bool = False) -> None:
        column_widget = ui.create_column_widget()
        super().__init__(CompositeWidgetBehavior(column_widget))
        self.property_changed_event = Event.Event()
        items = items or list()
        self.__items: typing.List[typing.Any] = list()
        self.on_selection_changed: typing.Optional[typing.Callable[[typing.AbstractSet[int]], None]] = None
        self.on_item_selected: typing.Optional[typing.Callable[[int], bool]] = None
        self.on_cancel: typing.Optional[typing.Callable[[], None]] = None
        self.on_item_handle_context_menu: typing.Optional[typing.Callable[..., bool]] = None  # used for declarative
        self.__items_binding: typing.Optional[Binding.Binding] = None
        self.__v_auto_resize = v_auto_resize
        self.on_escape_pressed : typing.Optional[typing.Callable[[], bool]] = None
        self.on_return_pressed : typing.Optional[typing.Callable[[], bool]] = None

        self.__selection = selection if selection else Selection.IndexedSelection(selection_style)

        def selection_changed() -> None:
            self.__current_index_binding_helper.value_changed(self.__selection.current_index or 0)
            on_selection_changed = self.on_selection_changed
            if callable(on_selection_changed):
                on_selection_changed(self.__selection.indexes)

        def handle_delegate_cancel() -> None:
            if callable(self.on_cancel):
                self.on_cancel()
            if callable(self.on_escape_pressed):
                self.on_escape_pressed()

        def handle_delegate_item_selected(index: int) -> None:
            if callable(self.on_item_selected):
                self.on_item_selected(index)
            if callable(self.on_return_pressed):
                self.on_return_pressed()

        self.__selection_changed_event_listener = self.__selection.changed_event.listen(selection_changed)
        self.__list_canvas_item_delegate = list_item_delegate
        self.__list_canvas_item_delegate.on_cancel = handle_delegate_cancel
        self.__list_canvas_item_delegate.on_item_selected = handle_delegate_item_selected
        self.__list_canvas_item = ListCanvasItem.ListCanvasItem(self.__list_canvas_item_delegate, self.__selection, 20)

        scroll_area_canvas_item = CanvasItem.ScrollAreaCanvasItem(self.__list_canvas_item)
        scroll_area_canvas_item.auto_resize_contents = True
        scroll_group_canvas_item = CanvasItem.CanvasItemComposition()
        if border_color is not None:
            scroll_group_canvas_item.border_color = border_color
        scroll_group_canvas_item.layout = CanvasItem.CanvasItemRowLayout()
        scroll_group_canvas_item.add_canvas_item(scroll_area_canvas_item)
        if v_scroll_enabled:
            scroll_bar_canvas_item = CanvasItem.ScrollBarCanvasItem(scroll_area_canvas_item)
            scroll_group_canvas_item.add_canvas_item(scroll_bar_canvas_item)

        canvas_widget = ui.create_canvas_widget(properties=properties)
        canvas_widget.canvas_item.add_canvas_item(scroll_group_canvas_item)

        column_widget.add(canvas_widget)

        self.__canvas_widget = canvas_widget

        def set_current_index(index: int) -> None:
            self.__selection.set(index)

        def validate_current_index(new_value: int, old_value: int) -> int:
            return new_value if new_value >= 0 and new_value < len(self.items) else 0

        def set_items(items: typing.Sequence[typing.Any]) -> None:
            self.__set_items(items)

        self.__current_index_binding_helper = UserInterface.BindablePropertyHelper[int](None, set_current_index, validate_current_index)
        self.__items_binding_helper = UserInterface.BindablePropertyHelper[typing.Sequence[typing.Any]](None, set_items)

        self.items = list(items) if items else list()
        self.current_index = 0

    def close(self) -> None:
        self.__selection_changed_event_listener.close()
        self.__selection_changed_event_listener = typing.cast(typing.Any, None)
        self.on_selection_changed = None
        self.__items_binding_helper.close()
        self.__items_binding_helper = typing.cast(typing.Any, None)
        self.__current_index_binding_helper.close()
        self.__current_index_binding_helper = typing.cast(typing.Any, None)
        self.__items = typing.cast(typing.Any, None)
        super().close()

    @property
    def focused(self) -> bool:
        return self.__canvas_widget.focused and self.__list_canvas_item.focused

    @focused.setter
    def focused(self, focused: bool) -> None:
        self.__list_canvas_item.request_focus()

    @property
    def wants_drag_events(self) -> bool:
        return self.__list_canvas_item.wants_drag_events

    @wants_drag_events.setter
    def wants_drag_events(self, value: bool) -> None:
        self.__list_canvas_item.wants_drag_events = value

    @property
    def items(self) -> typing.Sequence[typing.Any]:
        return self.__items_binding_helper.value

    @items.setter
    def items(self, items: typing.Sequence[typing.Any]) -> None:
        self.__items_binding_helper.value = items

    def __set_items(self, items: typing.Sequence[typing.Any]) -> None:
        self.__items = list(items)

        self.__list_canvas_item_delegate.items = self.__items
        self.__list_canvas_item.size_to_content()  # this will only configure the sizing, not actual size (until next layout)
        self.__list_canvas_item.update()

        if self.__v_auto_resize:
            # if v_auto_resize is True, ensure the canvas item resizes vertically to its content and the canvas widget
            # resizes vertically to the height of the canvas item content.

            new_sizing = self.__canvas_widget.canvas_item.sizing
            content_height = self.__list_canvas_item.sizing.maximum_height
            new_sizing = new_sizing.with_minimum_height(content_height)
            new_sizing = new_sizing.with_preferred_height(content_height)
            new_sizing = new_sizing.with_maximum_height(content_height)
            self.__canvas_widget.canvas_item.update_sizing(new_sizing)

            self.__canvas_widget.set_property("min-height", content_height)
            self.__canvas_widget.set_property("max-height", content_height)
            self.__canvas_widget.set_property("size-policy-vertical", "fixed")

        # setting items on the widget will not update items on the bound items since the list widget is merely a view

    def bind_items(self, binding: Binding.Binding) -> None:
        self.__items_binding_helper.bind_value(binding)

    def unbind_items(self) -> None:
        self.__items_binding_helper.unbind_value()

    @property
    def selected_items(self) -> typing.AbstractSet[int]:
        return self.__selection.indexes

    def set_selected_index(self, index: int) -> None:
        self.__selection.set(index)
        self.__list_canvas_item.make_selection_visible()

    def update(self) -> None:
        self.__list_canvas_item.update()

    @property
    def current_index(self) -> int:
        return self.__current_index_binding_helper.value

    @current_index.setter
    def current_index(self, index: int) -> None:
        self.__current_index_binding_helper.value = index

    def bind_current_index(self, binding: Binding.Binding) -> None:
        self.__current_index_binding_helper.bind_value(binding)

    def unbind_current_index(self) -> None:
        self.__current_index_binding_helper.unbind_value()


class StringListCanvasItemDelegate(ListCanvasItemDelegate):

    def __init__(self, item_getter: typing.Optional[typing.Callable[[typing.Any], str]] = None):
        super().__init__()
        self.__item_getter = item_getter or (lambda x: str(x))

    def close(self) -> None:
        self.__item_getter = typing.cast(typing.Any, None)

    def paint_item(self, drawing_context: DrawingContext.DrawingContext, display_item: typing.Any, rect: Geometry.IntRect, is_selected: bool) -> None:

        def notnone(s: typing.Any) -> str:
            return str(s) if s is not None else str()

        item_string = notnone(self.__item_getter(display_item) if callable(self.__item_getter) else display_item)

        with drawing_context.saver():
            drawing_context.fill_style = "#000"
            drawing_context.font = "12px"
            drawing_context.text_align = 'left'
            drawing_context.text_baseline = 'bottom'
            drawing_context.fill_text(item_string, rect.left + 4, rect.top + 20 - 4)


class StringListWidget(ListWidget):
    """A widget with a list in a scroll bar."""

    # TODO: arguments after ui should be keyword required. current arguments are for backwards compatibility.
    def __init__(self, ui: UserInterface.UserInterface, items: typing.Optional[typing.Sequence[typing.Any]] = None,
                 selection_style: typing.Optional[Selection.Style] = None,
                 item_getter: typing.Optional[typing.Callable[[typing.Any], str]] = None, *,
                 properties: typing.Optional[typing.Mapping[str, typing.Any]] = None,
                 selection: typing.Optional[Selection.IndexedSelection] = None,
                 border_color: typing.Optional[str] = None, v_scroll_enabled: bool = True,
                 v_auto_resize: bool = False) -> None:
        super().__init__(ui, StringListCanvasItemDelegate(item_getter), items=items, selection_style=selection_style,
                         properties=properties, selection=selection, border_color=border_color,
                         v_scroll_enabled=v_scroll_enabled, v_auto_resize=v_auto_resize)


class TableWidget(UserInterface.Widget):
    """A widget representing a table (column only)."""

    def __init__(self, ui: UserInterface.UserInterface,
                 create_list_item_widget: typing.Callable[[typing.Any], UserInterface.BoxWidget],
                 header_widget: typing.Optional[UserInterface.Widget] = None,
                 header_for_empty_list_widget: typing.Optional[UserInterface.Widget] = None) -> None:
        column_widget = ui.create_column_widget()
        super().__init__(CompositeWidgetBehavior(column_widget))
        self.__binding: typing.Optional[Binding.Binding] = None
        self.content_section = ui.create_column_widget()
        self.content_section.widget_id = "content_section"
        header_column = ui.create_column_widget()
        self.header_widget = header_widget
        self.header_for_empty_list_widget = header_for_empty_list_widget
        if self.header_widget:
            header_column.add(self.header_widget)
        if self.header_for_empty_list_widget:
            header_column.add(self.header_for_empty_list_widget)
        column_widget.add(header_column)
        content_column = ui.create_column_widget()
        content_column.add(self.content_section)
        content_column.add_stretch()
        column_widget.add(content_column)
        column_widget.add_stretch()
        self.create_list_item_widget = create_list_item_widget

    def close(self) -> None:
        if self.__binding:
            self.__binding.close()
            self.__binding = None
        self.clear_queued_tasks()
        self.content_section = typing.cast(typing.Any, None)
        self.header_widget = None
        self.header_for_empty_list_widget = None
        self.create_list_item_widget = typing.cast(typing.Any, None)
        super().close()

    @property
    def list_items(self) -> typing.Sequence[UserInterface.Widget]:
        return self.content_section.children

    @property
    def list_item_count(self) -> int:
        return self.content_section.child_count

    def insert_item(self, item: typing.Any, before_index: int) -> None:
        if callable(self.create_list_item_widget):  # item may be closed while this call is pending on main thread.
            item_row = self.create_list_item_widget(item)
            if self.content_section:
                self.content_section.insert(item_row, before_index)
                self.__sync_header()
            else:
                item_row.close()

    def remove_item(self, index: int) -> None:
        if self.content_section:
            self.content_section.remove(index)
            self.__sync_header()

    def remove_all_items(self) -> None:
        self.content_section.remove_all()
        self.__sync_header()

    def __sync_header(self) -> None:
        # select the right header item
        has_content = self.content_section.child_count > 0
        if self.header_widget:
            self.header_widget.visible = has_content
        if self.header_for_empty_list_widget:
            self.header_for_empty_list_widget.visible = not has_content


class TextPushButtonWidget(UserInterface.Widget):
    def __init__(self, ui: UserInterface.UserInterface, text: str) -> None:
        column_widget = ui.create_column_widget()
        super().__init__(CompositeWidgetBehavior(column_widget))
        self.on_button_clicked: typing.Optional[typing.Callable[[], None]] = None
        font = "normal 11px serif"
        font_metrics = ui.get_font_metrics(font, text)
        text_button_canvas_item = CanvasItem.TextButtonCanvasItem(text)
        text_button_canvas_item.update_sizing(text_button_canvas_item.sizing.with_fixed_size(Geometry.IntSize(height=font_metrics.height + 6, width=font_metrics.width + 6)))

        def button_clicked() -> None:
            if callable(self.on_button_clicked):
                self.on_button_clicked()

        text_button_canvas_item.on_button_clicked = button_clicked

        text_button_canvas_widget = ui.create_canvas_widget(properties={"height": 20, "width": 20})
        text_button_canvas_widget.canvas_item.add_canvas_item(text_button_canvas_item)
        # ugh. this is a partially working stop-gap when a canvas item is in a widget it will not get mouse exited reliably
        root_container = text_button_canvas_item.root_container
        if root_container:
            text_button_canvas_widget.on_mouse_exited = root_container.canvas_widget.on_mouse_exited

        column_widget.add(text_button_canvas_widget)


class ImageWidget(UserInterface.Widget):
    def __init__(self, ui: UserInterface.UserInterface,
                 bitmap: typing.Optional[Bitmap.Bitmap] = None,
                 properties: typing.Optional[typing.Mapping[str, typing.Any]] = None) -> None:
        column_widget = ui.create_column_widget(properties=properties)
        super().__init__(CompositeWidgetBehavior(column_widget))
        self.ui = ui
        self.on_clicked = None

        def button_clicked() -> None:
            if callable(self.on_clicked):
                self.on_clicked()

        self.__bitmap_canvas_item = CanvasItem.BitmapButtonCanvasItem(bitmap)
        self.__bitmap_canvas_item.on_button_clicked = button_clicked
        bitmap_canvas_widget = self.ui.create_canvas_widget()
        bitmap_canvas_widget.canvas_item.add_canvas_item(self.__bitmap_canvas_item)
        column_widget.add(bitmap_canvas_widget)

        def set_image(value: typing.Optional[Bitmap.BitmapOrArray]) -> None:
            self.__bitmap_canvas_item.bitmap = Bitmap.promote_bitmap(value)
            self.__bitmap_canvas_item.size_to_content(ui.get_font_metrics)

        self.__image_binding_helper = UserInterface.BindablePropertyHelper[typing.Optional[Bitmap.BitmapOrArray]](None, set_image, None, typing.cast(typing.Any, Bitmap.bitmap_or_array_equal))

        self.image = bitmap

    def close(self) -> None:
        self.__image_binding_helper.close()
        self.__image_binding_helper = typing.cast(typing.Any, None)
        super().close()

    @property
    def image(self) -> typing.Optional[Bitmap.Bitmap]:
        return Bitmap.promote_bitmap(self.__image_binding_helper.value)

    @image.setter
    def image(self, bitmap: typing.Optional[Bitmap.BitmapOrArray]) -> None:
        self.__image_binding_helper.value = bitmap

    def bind_image(self, binding: Binding.Binding) -> None:
        self.__image_binding_helper.bind_value(binding)

    def unbind_image(self) -> None:
        self.__image_binding_helper.unbind_value()

    @property
    def background_color(self) -> typing.Optional[typing.Union[str, DrawingContext.LinearGradient]]:
        return self.__bitmap_canvas_item.background_color

    @background_color.setter
    def background_color(self, background_color: typing.Optional[typing.Union[str, DrawingContext.LinearGradient]]) -> None:
        self.__bitmap_canvas_item.background_color = background_color

    @property
    def border_color(self) -> typing.Optional[str]:
        return self.__bitmap_canvas_item.border_color

    @border_color.setter
    def border_color(self, border_color: typing.Optional[str]) -> None:
        self.__bitmap_canvas_item.border_color = border_color


class ColorButtonCell(CanvasItem.Cell):

    def __init__(self, color: typing.Optional[str]):
        super().__init__(padding=Geometry.IntSize(4, 4))
        self.__color = color

    @property
    def color(self) -> typing.Optional[str]:
        return self.__color

    @color.setter
    def color(self, value: typing.Optional[str]) -> None:
        self.__color = value

    def _size_to_content(self, get_font_metrics_fn: UserInterface.MeasureTextFn) -> Geometry.IntSize:
        """ Size the canvas item to the text content without padding."""
        return Geometry.IntSize(height=30, width=44)

    def _paint_cell(self, drawing_context: DrawingContext.DrawingContext, rect: Geometry.FloatRect, style: typing.Set[str]) -> None:
        # style: "disabled" (default is enabled)

        drawing_context.begin_path()
        drawing_context.rect(rect.left, rect.top, rect.width, rect.height)
        drawing_context.fill_style = "#BBB"
        drawing_context.fill()

        inset_rect = rect.inset(4, 4)

        drawing_context.begin_path()
        drawing_context.rect(inset_rect.left, inset_rect.top, inset_rect.width, inset_rect.height)
        drawing_context.close_path()
        drawing_context.fill_style = self.__color
        drawing_context.fill()

        drawing_context.begin_path()
        drawing_context.move_to(inset_rect.right, inset_rect.top)
        drawing_context.line_to(inset_rect.right, inset_rect.bottom)
        drawing_context.line_to(inset_rect.left, inset_rect.bottom)
        drawing_context.close_path()
        drawing_context.fill_style = DrawingContext.color_without_alpha(self.__color)
        drawing_context.fill()

        drawing_context.begin_path()
        drawing_context.rect(inset_rect.left, inset_rect.top, inset_rect.width, inset_rect.height)
        drawing_context.stroke_style = "#454545"
        drawing_context.stroke()


class ColorButtonCanvasItem(CanvasItem.CellCanvasItem):

    def __init__(self, color: typing.Optional[str]):
        super().__init__()
        self.color_button_cell = ColorButtonCell(color)
        self.cell = self.color_button_cell
        self.wants_mouse_events = True
        self.on_button_clicked: typing.Optional[typing.Callable[[], None]] = None

    def close(self) -> None:
        self.on_button_clicked = None
        super().close()

    def mouse_entered(self) -> bool:
        self._mouse_inside = True
        return True

    def mouse_exited(self) -> bool:
        self._mouse_inside = False
        return True

    def mouse_pressed(self, x: int, y: int, modifiers: UserInterface.KeyboardModifiers) -> bool:
        self._mouse_pressed = True
        return True

    def mouse_released(self, x: int, y: int, modifiers: UserInterface.KeyboardModifiers) -> bool:
        self._mouse_pressed = False
        return True

    def mouse_clicked(self, x: int, y: int, modifiers: UserInterface.KeyboardModifiers) -> bool:
        if self.enabled:
            if self.on_button_clicked:
                self.on_button_clicked()
        return True


class ColorPushButtonWidget(UserInterface.Widget):
    def __init__(self, ui: UserInterface.UserInterface, color: typing.Optional[str] = None):
        column_widget = ui.create_column_widget()
        super().__init__(CompositeWidgetBehavior(column_widget))

        self.on_color_changed: typing.Optional[typing.Callable[[typing.Optional[str]], None]] = None

        color_button_canvas_item = ColorButtonCanvasItem(color)
        color_button_canvas_item.update_sizing(color_button_canvas_item.sizing.with_fixed_size(Geometry.IntSize(height=30, width=44)))

        def button_clicked() -> None:
            start_color = self.color or "rgba(255, 255, 255, 0.0)"
            color = ui.get_color_dialog(_("Select Color"), start_color, True)
            if color != start_color:
                self.color = color
                if callable(self.on_color_changed):
                    self.on_color_changed(self.color)

        color_button_canvas_item.on_button_clicked = button_clicked

        color_button_canvas_widget = ui.create_canvas_widget(properties={"height": 30, "width": 44})
        color_button_canvas_widget.canvas_item.add_canvas_item(color_button_canvas_item)
        # ugh. this is a partially working stop-gap when a canvas item is in a widget it will not get mouse exited reliably
        root_container = color_button_canvas_item.root_container
        if root_container:
            color_button_canvas_widget.on_mouse_exited = root_container.canvas_widget.on_mouse_exited

        self.__color_button_canvas_item = color_button_canvas_item

        column_widget.add(color_button_canvas_widget)

        def set_color(color: typing.Optional[str]) -> None:
            self.__color_button_canvas_item.color_button_cell.color = color
            self.__color_button_canvas_item.update()

        self.__color_binding_helper = UserInterface.BindablePropertyHelper[typing.Optional[str]](None, set_color)

        self.color = color

    def close(self) -> None:
        self.__color_binding_helper.close()
        self.__color_binding_helper = typing.cast(typing.Any, None)
        super().close()

    @property
    def color(self) -> typing.Optional[str]:
        return self.__color_binding_helper.value

    @color.setter
    def color(self, color: typing.Optional[str]) -> None:
        self.__color_binding_helper.value = color

    def bind_color(self, binding: Binding.Binding) -> None:
        self.__color_binding_helper.bind_value(binding)

    def unbind_color(self) -> None:
        self.__color_binding_helper.unbind_value()


class IconRadioButtonWidget(UserInterface.Widget):

    def __init__(self, ui: UserInterface.UserInterface, properties: typing.Optional[typing.Mapping[str, typing.Any]] = None):
        column_widget = ui.create_column_widget(properties=properties)
        # must be available for enabled property in super class. needs refactoring.
        self.__bitmap_canvas_item = CanvasItem.BitmapButtonCanvasItem(None, border_color="#CCC")
        super().__init__(CompositeWidgetBehavior(column_widget))
        self.ui = ui
        self.on_clicked: typing.Optional[typing.Callable[[], None]] = None
        self.__bitmap_canvas_item.on_button_clicked = self.__handle_clicked
        self.__value: typing.Optional[int] = None
        self.__enabled = True
        self.__checked = False
        bitmap_canvas_widget = self.ui.create_canvas_widget()
        bitmap_canvas_widget.canvas_item.add_canvas_item(self.__bitmap_canvas_item)
        column_widget.add(bitmap_canvas_widget)

        def set_group_value(group_value: typing.Optional[int]) -> None:
            self.checked = group_value == self.__value

        def set_icon(value: typing.Optional[Bitmap.BitmapOrArray]) -> None:
            self.__bitmap_canvas_item.bitmap = Bitmap.promote_bitmap(value)

        self.__group_value_binding_helper = UserInterface.BindablePropertyHelper[typing.Optional[typing.Optional[int]]](None, set_group_value)
        self.__icon_binding_helper = UserInterface.BindablePropertyHelper[typing.Optional[Bitmap.BitmapOrArray]](None, set_icon, None, typing.cast(typing.Any, Bitmap.bitmap_or_array_equal))

    def close(self) -> None:
        self.__group_value_binding_helper.close()
        self.__group_value_binding_helper = typing.cast(typing.Any, None)
        self.__icon_binding_helper.close()
        self.__icon_binding_helper = typing.cast(typing.Any, None)
        super().close()

    def __handle_clicked(self) -> None:
        if self.__value is not None:
            self.group_value = self.__value
        if callable(self.on_clicked):
            self.on_clicked()

    @property
    def enabled(self) -> bool:
        return self.__bitmap_canvas_item.enabled

    @enabled.setter
    def enabled(self, value: bool) -> None:
        self.__bitmap_canvas_item.enabled = value

    @property
    def checked(self) -> bool:
        return self.__bitmap_canvas_item.checked

    @checked.setter
    def checked(self, value: bool) -> None:
        self.__bitmap_canvas_item.checked = value

    @property
    def icon(self) -> typing.Optional[Bitmap.Bitmap]:
        return Bitmap.promote_bitmap(self.__icon_binding_helper.value)

    @icon.setter
    def icon(self, bitmap: typing.Optional[Bitmap.BitmapOrArray]) -> None:
        self.__icon_binding_helper.value = bitmap

    @property
    def value(self) -> typing.Optional[int]:
        return self.__value

    @value.setter
    def value(self, value: typing.Optional[int]) -> None:
        self.__value = value
        self.checked = self.group_value == self.__value

    @property
    def group_value(self) -> typing.Optional[int]:
        return self.__group_value_binding_helper.value

    @group_value.setter
    def group_value(self, group_value: typing.Optional[int]) -> None:
        self.__group_value_binding_helper.value = group_value

    def bind_group_value(self, binding: Binding.Binding) -> None:
        self.__group_value_binding_helper.bind_value(binding)

    def unbind_group_value(self) -> None:
        self.__group_value_binding_helper.unbind_value()

    def bind_icon(self, binding: Binding.Binding) -> None:
        self.__icon_binding_helper.bind_value(binding)

    def unbind_icon(self) -> None:
        self.__icon_binding_helper.unbind_value()


class CompositeWidgetBase(UserInterface.Widget):
    # deprecated - no longer required - just pass the behavior directly to the widget

    def __init__(self, content_widget: UserInterface.Widget) -> None:
        super().__init__(CompositeWidgetBehavior(content_widget))



from nion.ui import Declarative
from nion.utils import Model


class Handler(Declarative.Handler):

    slider_value_model = Model.PropertyModel(50)

    def reset(self, widget: Declarative.UIWidget) -> None:
        self.slider_value_model.value = 50


def construct_ui(u: Declarative.DeclarativeUI) -> Declarative.UIDescription:
    slider = u.create_slider(value="@binding(slider_value_model.value)")
    progress_bar = u.create_progress_bar(value="@binding(slider_value_model.value)")
    label = u.create_label(text="@binding(slider_value_model.value)")
    button = u.create_push_button(text="Reset to 50", on_clicked="reset")
    return u.create_column(progress_bar, slider, label, button, spacing=8)



import numpy
import numpy.typing
import typing

from nion.ui import Declarative
from nion.utils import Event
from nion.utils import Model

IconType = numpy.typing.NDArray[typing.Any]


def make_icon(w: int, h: int) -> IconType:
    # make bitmap_data (random static) for icon push button
    bitmap: IconType = numpy.zeros((h, w, 4), numpy.uint8)
    bitmap[..., 0] = (numpy.random.randn(h, w) * 255).astype(numpy.uint8)  # blue
    bitmap[..., 1] = (numpy.random.randn(h, w) * 255).astype(numpy.uint8)  # green
    bitmap[..., 2] = (numpy.random.randn(h, w) * 255).astype(numpy.uint8)  # red
    bitmap[..., 3] = 255
    return bitmap.view(numpy.uint32).reshape(bitmap.shape[:-1])


class Handler(Declarative.Handler):

    def __init__(self) -> None:
        super().__init__()
        self.icon1 =  make_icon(12, 12)
        self.icon2 =  make_icon(48, 36)
        self.image_model = Model.PropertyModel(make_icon(16, 16), typing.cast(typing.Callable[[typing.Optional[IconType], typing.Optional[IconType]], bool], numpy.array_equal))
        self.property_changed_event = Event.Event()

    def click(self, widget: Declarative.UIWidget) -> None:
        self.icon1 =  make_icon(12, 12)
        self.property_changed_event.fire("icon1")

    def image_click(self, widget: Declarative.UIWidget) -> None:
        self.image_model.value = make_icon(16, 16)


def construct_ui(u: Declarative.DeclarativeUI) -> Declarative.UIDescription:
    text_only_button = u.create_push_button(text="Change Button Icon", on_clicked="click")
    icon_only_button = u.create_push_button(icon="@binding(icon2)")
    text_and_icon_button = u.create_push_button(text="Text", icon="@binding(icon1)")
    image_button = u.create_image(image="@binding(image_model.value)", height=16, width=16, on_clicked="image_click")

    hellos = u.create_column(
        u.create_row(text_only_button, u.create_stretch()),
        u.create_row(icon_only_button, u.create_stretch()),
        u.create_row(text_and_icon_button, u.create_stretch()),
        u.create_row(u.create_label(text="Image (click me):"), image_button, u.create_stretch(), spacing=8),
        u.create_stretch(),
        spacing=12)

    return hellos



from __future__ import annotations

import typing

from nion.ui import Declarative
from nion.ui import UserInterface
from nion.utils import Model


class Handler(Declarative.Handler):
    def __init__(self) -> None:
        super().__init__()
        self.color_line_edit: typing.Optional[UserInterface.LineEditWidget] = None
        self.model = Model.PropertyModel("red")

    def color_updated(self, widget: Declarative.UIWidget, text: str) -> None:
        print(f"Color changed: {widget} {text}")

    def return_pressed(self, widget: Declarative.UIWidget) -> bool:
        assert self.color_line_edit
        self.color_line_edit.select_all()
        return False  # required so that editing_finished event is triggered

    def key_pressed(self, widget: Declarative.UIWidget, key: UserInterface.Key) -> bool:
        if key.text == "*":
            return True
        return False


def construct_ui(u: Declarative.DeclarativeUI) -> Declarative.UIDescription:
    field_label = u.create_label(text="Favorite Color?")
    field_line_edit = u.create_line_edit(placeholder_text="Color", name="color_line_edit", on_editing_finished="color_updated", on_return_pressed="return_pressed", on_key_pressed="key_pressed")
    field_line_edit2 = u.create_line_edit(text="@binding(model.value)")
    return u.create_column(
        u.create_row(field_label, field_line_edit, spacing=8),
        u.create_row(u.create_label(text="Another Color?"), field_line_edit2, u.create_stretch(), spacing=8),
        u.create_stretch(),
        spacing=8,
    )



import typing

from nion.ui import Declarative
from nion.ui import Widgets
from nion.utils import Model


class Handler(Declarative.Handler):
    def __init__(self) -> None:
        super().__init__()
        self.expanded_model0 = Model.PropertyModel(False)
        self.expanded_model1 = Model.PropertyModel(True)
        self.expanded_model2 = Model.PropertyModel(False)
        self.section: typing.Optional[Widgets.SectionWidget] = None

    def switch3(self, widget: Declarative.UIWidget) -> None:
        self.expanded_model0.value = False
        self.expanded_model1.value = False
        self.expanded_model2.value = True
        if self.section:
            self.section.title = "Tres"


def construct_ui(u: Declarative.DeclarativeUI) -> Declarative.UIDescription:

    def create_section(tab_label: str, text: str, check_text: str, button_text: str, index: int) -> Declarative.UIDescription:
        label = u.create_label(text=text)
        check_box = u.create_check_box(text=check_text)
        button = u.create_push_button(text=button_text)
        return u.create_section(u.create_column(label, check_box, button, spacing=8, margin=4), title=tab_label, expanded=f"@binding(expanded_model{index}.value)", name="section")

    section0 = create_section("First", "ONE", "Check 1", "Push ONE", 0)
    section1 = create_section("Second", "TWO", "Check 2", "Push TWO", 1)
    section2 = create_section("Third", "THREE", "Check 3", "Push THREE", 2)

    button = u.create_push_button(text="3", on_clicked="switch3")

    return u.create_column(section0, section1, section2, button)



from nion.ui import Declarative
from nion.utils import Model


class Handler(Declarative.Handler):
    slider_value_model = Model.PropertyModel(50)

    def reset(self, widget: Declarative.UIWidget) -> None:
        self.slider_value_model.value = 50


def construct_ui(u: Declarative.DeclarativeUI) -> Declarative.UIDescription:
    label = u.create_label(text="@binding(slider_value_model.value)")
    button = u.create_push_button(text="Reset to 50", on_clicked="reset")
    content = u.create_column(label, button, spacing=8)
    left = u.create_label(text="LEFT")
    right = u.create_label(text="RIGHT")
    group_row = u.create_row(left, u.create_stretch(), right, spacing=8)
    status_bar = u.create_group(group_row)
    return u.create_column(content, u.create_stretch(), status_bar, spacing=8)



# standard libraries
import contextlib
import logging
import typing
import unittest

# third party libraries
# None

# local libraries
from nion.ui import CanvasItem
from nion.ui import TestUI
from nion.ui import UserInterface
from nion.utils import Geometry


class TestCanvasItemClass(unittest.TestCase):

    def setUp(self) -> None:
        CanvasItem._threaded_rendering_enabled = False

    def tearDown(self) -> None:
        pass

    def test_add_item_to_string_list_widget_causes_container_to_relayout(self) -> None:
        # ugly type casting
        from nion.ui import Widgets
        ui = TestUI.UserInterface()
        widget = Widgets.StringListWidget(ui)
        with contextlib.closing(widget):
            canvas_item = typing.cast(CanvasItem.CanvasItemComposition, typing.cast(UserInterface.CanvasWidget, typing.cast(UserInterface.BoxWidget, widget.content_widget).children[0]).canvas_item)
            canvas_item.update_layout(Geometry.IntPoint(x=0, y=0), Geometry.IntSize(width=300, height=200), immediate=True)
            scroll_area_canvas_item = typing.cast(CanvasItem.ScrollAreaCanvasItem, typing.cast(CanvasItem.CanvasItemComposition, canvas_item.canvas_items[0]).canvas_items[0])
            canvas_item.layout_immediate(Geometry.IntSize(width=300, height=200))
            # check assumptions
            scroll_canvas_rect = scroll_area_canvas_item.canvas_rect or Geometry.IntRect.empty_rect()
            scroll_content = scroll_area_canvas_item.content
            assert scroll_content
            self.assertEqual(scroll_canvas_rect.height, 200)
            scroll_content_rect = scroll_content.canvas_rect or Geometry.IntRect.empty_rect()
            self.assertEqual(scroll_content_rect.height, 0)
            # add item
            self.assertFalse(canvas_item._needs_layout_for_testing)
            widget.items = ["abc"]
            # self.assertTrue(canvas_item._needs_layout_for_testing)
            # check that column was laid out again
            canvas_item.layout_immediate(Geometry.IntSize(width=300, height=200), force=False)
            scroll_canvas_rect = scroll_area_canvas_item.canvas_rect or Geometry.IntRect.empty_rect()
            scroll_content = scroll_area_canvas_item.content
            assert scroll_content
            scroll_content_rect = scroll_content.canvas_rect or Geometry.IntRect.empty_rect()
            self.assertEqual(scroll_canvas_rect.height, 200)
            self.assertEqual(scroll_content_rect.height, 20)


if __name__ == '__main__':
    logging.getLogger().setLevel(logging.DEBUG)
    unittest.main()



from nion.ui import Declarative
from nion.utils import Model


class Handler(Declarative.Handler):
    stack_index_model = Model.PropertyModel(1)


def construct_ui(u: Declarative.DeclarativeUI) -> Declarative.UIDescription:
    stack0 = u.create_column(u.create_label(text="111"), u.create_label(text="ONE"))
    stack1 = u.create_column(u.create_label(text="222"), u.create_label(text="TWO"))
    stack2 = u.create_column(u.create_label(text="333"), u.create_label(text="THREE"))
    stack = u.create_stack(stack0, stack1, stack2, current_index="@binding(stack_index_model.value)")
    chooser = u.create_combo_box(items=["One", "Two", "Three"], current_index="@binding(stack_index_model.value)")
    return u.create_column(stack, chooser, spacing=8)



"""Display a tree of drawable cells.
"""

from __future__ import annotations

# standard libraries
import copy
import dataclasses
import functools
import json
import threading
import typing

# third party libraries
# none

# local libraries
from nion.ui import CanvasItem
from nion.ui import UserInterface
from nion.utils import Geometry

_ValuePath = typing.Sequence[typing.Union[int, str]]


@dataclasses.dataclass
class TreeItem:
    canvas_item: CanvasItem.AbstractCanvasItem
    item_type: str
    is_expanded: bool
    value_path: _ValuePath


class TreeCanvasItemDelegate(typing.Protocol):

    def build_items(self, get_font_metrics_fn: typing.Callable[[str, str], UserInterface.FontMetrics],
                    item_width: typing.Optional[int]) -> typing.Sequence[TreeItem]: ...

    def toggle_is_expanded(self, value_path_key: str) -> None: ...


class TreeCanvasItem(CanvasItem.CanvasItemComposition):
    """
    Takes a delegate that supports the following properties, methods, and optional methods:

    Properties:
        None

    Methods:
        toggle_is_expanded(value_path) -> None
        build_items(get_font_metrics_fn, item_width) -> CanvasItem

    Optional methods:
        None

    Call reconstruct when data or selection changes.
    """

    def __init__(self, get_font_metrics_fn: typing.Callable[[str, str], UserInterface.FontMetrics], delegate: TreeCanvasItemDelegate) -> None:
        super().__init__()
        self.__get_font_metrics_fn = get_font_metrics_fn
        self.__delegate = delegate
        # configure super
        self.wants_mouse_events = True
        self.focusable = True
        # internal variables
        self.__mouse_pressed = False
        self.__mouse_index: typing.Optional[int] = None
        self.__mouse_position: typing.Optional[Geometry.IntPoint] = None
        self.__mouse_dragging = False
        self.__mouse_item: typing.Optional[_ValuePath] = None
        self.__selected_value_paths: typing.Set[str] = set()
        self.layout = CanvasItem.CanvasItemColumnLayout()
        self.on_content_height_changed: typing.Optional[typing.Callable[[int], None]] = None
        self.on_reconstruct: typing.Optional[typing.Callable[[], None]] = None

    def close(self) -> None:
        self.on_content_height_changed = None
        super().close()

    def __is_selected(self, value_path: _ValuePath) -> bool:
        return json.dumps(value_path) in self.__selected_value_paths

    def reconstruct(self) -> None:
        for canvas_item in copy.copy(self.canvas_items):
            self._remove_canvas_item(canvas_item)
        indent_size = 16
        canvas_bounds = self.canvas_bounds
        item_width = int(canvas_bounds.width) if canvas_bounds else None
        canvas_height = 0
        ITEM_HEIGHT = 18
        for tree_item in self.__delegate.build_items(self.__get_font_metrics_fn, item_width):
            indent = (len(tree_item.value_path) - 1) * indent_size
            item_row = CanvasItem.CanvasItemComposition()
            item_row.update_sizing(item_row.sizing.with_fixed_height(ITEM_HEIGHT))
            item_row.layout = CanvasItem.CanvasItemRowLayout()
            item_row.add_spacing(indent)
            if tree_item.item_type == "parent":
                twist_down_canvas_item = CanvasItem.TwistDownCanvasItem()
                twist_down_canvas_item.update_sizing(twist_down_canvas_item.sizing.with_fixed_size(Geometry.IntSize(height=ITEM_HEIGHT, width=16)))
                twist_down_canvas_item.checked = tree_item.is_expanded

                def twist_down_clicked(toggle_value_path: _ValuePath) -> None:
                    self.__toggle_is_expanded(toggle_value_path)

                twist_down_canvas_item.on_button_clicked = functools.partial(twist_down_clicked, tree_item.value_path)
                item_row.add_canvas_item(twist_down_canvas_item)
            else:
                item_row.add_spacing(indent_size)
            item_row.add_canvas_item(tree_item.canvas_item)
            item_row.add_stretch()
            setattr(item_row, "value_path", tree_item.value_path)
            self.add_canvas_item(item_row)
            canvas_height += ITEM_HEIGHT
        self.update()
        if callable(self.on_content_height_changed):
            self.on_content_height_changed(canvas_height)

    def __set_selection(self, value_path: _ValuePath) -> None:
        self.__selected_value_paths.clear()
        self.__selected_value_paths.add(json.dumps(value_path))

    def __extend_selection(self, value_path: _ValuePath) -> None:
        pass

    def __toggle_selection(self, value_path: _ValuePath) -> None:
        value_path_key = json.dumps(value_path)
        if value_path_key in self.__selected_value_paths:
            self.__selected_value_paths.remove(value_path_key)
        else:
            self.__selected_value_paths.add(value_path_key)

    def __toggle_is_expanded(self, value_path: _ValuePath) -> None:
        value_path_key = json.dumps(value_path)
        self.__delegate.toggle_is_expanded(value_path_key)
        if callable(self.on_reconstruct):
            self.on_reconstruct()
        else:
            self.reconstruct()
        self.update()

    def __context_menu_event(self, value_path: typing.Optional[_ValuePath], x: int, y: int, gx: int, gy: int) -> bool:
        return False

    def __drag_started(self, value_path: typing.Optional[_ValuePath], x: int, y: int, modifiers: UserInterface.KeyboardModifiers) -> None:
        pass

    def __delete(self) -> None:
        pass

    def __adjust_selection(self, action: str, extend: bool) -> None:
        pass

    def __value_path_at_point(self, p: Geometry.IntPoint) -> typing.Optional[_ValuePath]:
        for canvas_item in self.canvas_items_at_point(p.x, p.y):
            if hasattr(canvas_item, "value_path"):
                return typing.cast(_ValuePath, getattr(canvas_item, "value_path"))
        return None

    def context_menu_event(self, x: int, y: int, gx: int, gy: int) -> bool:
        p = Geometry.IntPoint(y=y, x=x)
        value_path = self.__value_path_at_point(p)
        if value_path:
            if not self.__is_selected(value_path):
                self.__set_selection(value_path)
            return self.__context_menu_event(value_path, x, y, gx, gy)
        return self.__context_menu_event(None, x, y, gx, gy)

    def mouse_pressed(self, x: int, y: int, modifiers: UserInterface.KeyboardModifiers) -> bool:
        p = Geometry.IntPoint(y=y, x=x)
        value_path = self.__value_path_at_point(p)
        if value_path:
            if modifiers.shift:
                self.__extend_selection(value_path)
            elif modifiers.control:
                self.__toggle_selection(value_path)
            else:
                self.__set_selection(value_path)
                self.__mouse_pressed = True
                self.__mouse_position = Geometry.IntPoint(y=y, x=x)
                self.__mouse_item = value_path
            return True
        return super().mouse_pressed(x, y, modifiers)

    def mouse_released(self, x: int, y: int, modifiers: UserInterface.KeyboardModifiers) -> bool:
        if self.__mouse_pressed:
            # double check whether mouse_released has been called explicitly as part of a drag.
            # see https://bugreports.qt.io/browse/QTBUG-40733
            pass  # leave this here for future reference
        self.__mouse_pressed = False
        self.__mouse_item = None
        self.__mouse_position = None
        self.__mouse_dragging = False
        return True

    def mouse_position_changed(self, x: int, y: int, modifiers: UserInterface.KeyboardModifiers) -> bool:
        if self.__mouse_pressed and self.__mouse_position:
            mouse_position_f = self.__mouse_position.to_float_point()
            point_f = Geometry.FloatPoint(y=y, x=x)
            if not self.__mouse_dragging and Geometry.distance(mouse_position_f, point_f) > 8:
                self.__mouse_dragging = True
                self.__drag_started(self.__mouse_item, x, y, modifiers)
                # once a drag starts, mouse release will not be called; call it here instead
                self.mouse_released(x, y, modifiers)
                return True
        return super().mouse_position_changed(x, y, modifiers)

    def key_pressed(self, key: UserInterface.Key) -> bool:
        if key.is_delete:
            self.__delete()
            return True
        if key.is_up_arrow:
            self.__adjust_selection("up", key.modifiers.shift)
            return True
        if key.is_down_arrow:
            self.__adjust_selection("down", key.modifiers.shift)
            return True
        return super().key_pressed(key)

    def handle_select_all(self) -> bool:
        self.__adjust_selection("all", False)
        return True



# standard libraries
import gettext
import typing

# third party libraries
# None

# local libraries
from nion.ui import Application
from nion.ui import Declarative

if typing.TYPE_CHECKING:
    from nion.ui import UserInterface


_ = gettext.gettext


# user program below

class Handler(Declarative.Handler):
    def __init__(self) -> None:
        super().__init__()
        self.label_item: typing.Optional[UserInterface.LabelWidget] = None
        self.click_count = 0

    def button_clicked(self, widget: Declarative.UIWidget) -> None:
        assert self.label_item
        self.click_count += 1
        self.label_item.text = _("Clicked") + " " + str(self.click_count)


def main(args: typing.Sequence[typing.Any], bootstrap_args: typing.Mapping[str, typing.Any]) -> Application.BaseApplication:
    ui = Declarative.DeclarativeUI()
    button = ui.create_push_button(text=_("Hello World"), on_clicked="button_clicked")
    label = ui.create_label(name="label_item", text=_("Not Clicked"))
    column = ui.create_column(button, label, spacing=8)
    window = ui.create_window(column, title=_("Hello World"), margin=12)
    return Application.run_window(args, bootstrap_args, window, Handler())



from __future__ import annotations

# standard libraries
import typing

# third party libraries
import numpy

# local libraries
from nion.ui import DrawingContext
from nion.utils import Geometry


class Bitmap:
    def __init__(self, *, rgba_bitmap_data: typing.Optional[DrawingContext.RGBA32Type] = None, shape: typing.Optional[Geometry.IntSize] = None) -> None:
        self.__rgba_bitmap_data = rgba_bitmap_data
        self.__shape = shape

    @property
    def rgba_bitmap_data(self) -> typing.Optional[DrawingContext.RGBA32Type]:
        return self.__rgba_bitmap_data

    @rgba_bitmap_data.setter
    def rgba_bitmap_data(self, rgba_bitmap_data: typing.Optional[DrawingContext.RGBA32Type]) -> None:
        self.__rgba_bitmap_data = rgba_bitmap_data

    @property
    def shape(self) -> typing.Optional[Geometry.IntSize]:
        return self.__shape

    @shape.setter
    def shape(self, shape: typing.Optional[Geometry.IntSize]) -> None:
        self.__shape = shape

    @property
    def computed_shape(self) -> Geometry.IntSize:
        return self.__shape if self.__shape else (Geometry.IntSize.make(typing.cast(Geometry.IntSizeTuple, self.__rgba_bitmap_data.shape)) if self.__rgba_bitmap_data is not None else Geometry.IntSize())


def promote_bitmap(bitmap_or_array: typing.Optional[BitmapOrArray]) -> typing.Optional[Bitmap]:
    if isinstance(bitmap_or_array, Bitmap):
        return bitmap_or_array
    if isinstance(bitmap_or_array, numpy.ndarray) and len(bitmap_or_array.shape) == 2 and bitmap_or_array.dtype == numpy.uint32:
        return Bitmap(rgba_bitmap_data=bitmap_or_array)
    return None


BitmapOrArray = typing.Union[Bitmap, DrawingContext.RGBA32Type]


def bitmap_or_array_equal(b1: typing.Any, b2: typing.Any) -> bool:
    bitmap1 = promote_bitmap(b1)
    bitmap2 = promote_bitmap(b2)
    if not bitmap1 and not bitmap2:
        return True
    if bitmap1 and bitmap2 and bitmap1.rgba_bitmap_data is not None and bitmap2.rgba_bitmap_data is not None:
        return numpy.array_equal(bitmap1.rgba_bitmap_data, bitmap2.rgba_bitmap_data)
    return False
